#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2012-2016 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from __future__ import unicode_literals

from pyannote.core import Annotation, Timeline
from .base import BaseMetric
from .utils import UEMSupportMixin

DER_NAME = 'detection error rate'
DER_TOTAL = 'total'
DER_FALSE_ALARM = 'false alarm'
DER_MISS = 'miss'


class DetectionErrorRate(UEMSupportMixin, BaseMetric):
    """Detection error rate

    This metric can be used to evaluate binary classification tasks such as
    speech activity detection, for instance. It can be applied on Timeline
    or Annotation instances. Inputs are expected to only contain segments (or
    tracks) corresponding to the positive class (e.g. speech regions). Gaps in
    the inputs considered as the negative class (e.g. non-speech regions).

    It is computed as (fa + miss) / total, where fa is the duration of false
    alarm (e.g. non-speech classified as speech), miss is the duration of
    missed detection (e.g. speech classified as non-speech), and total is the
    total duration of the positive class in the reference.

    Parameters
    ----------
    collar : float, optional
        Duration (in seconds) of collars removed from evaluation around
        boundaries of reference segments (one half before, one half after).
    """

    @classmethod
    def metric_name(cls):
        return DER_NAME

    @classmethod
    def metric_components(cls):
        return [DER_FALSE_ALARM, DER_MISS, DER_TOTAL]

    def __init__(self, collar=0., **kwargs):
        super(DetectionErrorRate, self).__init__()
        self.collar = collar

    def _preprocess(self, reference, uem=None):

        if isinstance(reference, Annotation):
            positive = reference.get_timeline().coverage()
        elif isinstance(reference, Timeline):
            positive = reference.coverage()
        else:
            raise NotImplementedError(
                'Only "Timeline" and "Annotation" are supported.')

        annotation = Annotation()

        for segment in positive:
            annotation[segment] = '+'
        for segment in positive.gaps(focus=uem):
            annotation[segment] = '-'

        return annotation

    def _confusion(self, reference, hypothesis, uem=None, **kwargs):
        """Confusion matrix"""

        reference, hypothesis = self.uemify(
            self._preprocess(reference, uem=uem),
            self._preprocess(hypothesis, uem=uem),
            uem=uem, collar=self.collar)

        return reference * hypothesis

    def _get_details(self, reference, hypothesis, uem=None, **kwargs):

        detail = self._init_details()

        confusion = self._confusion(reference, hypothesis, uem=uem)

        try:
            true_positive = confusion.loc['+', '+'].item()
        except KeyError as e:
            true_positive = 0.

        try:
            false_positive = confusion.loc['-', '+'].item()
        except KeyError as e:
            false_positive = 0.

        try:
            false_negative = confusion.loc['+', '-'].item()
        except KeyError as e:
            false_negative = 0.

        detail[DER_MISS] = false_negative
        detail[DER_FALSE_ALARM] = false_positive
        detail[DER_TOTAL] = true_positive + false_negative

        return detail

    def _get_rate(self, detail):
        error = 1. * (detail[DER_FALSE_ALARM] + detail[DER_MISS])
        total = 1. * detail[DER_TOTAL]
        if total == 0.:
            if error == 0:
                return 0.
            else:
                return 1.
        else:
            return error / total

    def _pretty(self, detail):
        string = ""
        string += "  - duration: %.2f seconds\n" % (detail[DER_TOTAL])
        string += "  - miss: %.2f seconds\n" % (detail[DER_MISS])
        string += "  - false alarm: %.2f seconds\n" % (detail[DER_FALSE_ALARM])
        string += "  - %s: %.2f %%\n" % (self.name, 100 * detail[self.name])
        return string

ACCURACY_NAME = 'detection accuracy'
ACCURACY_TRUE_POSITIVE = 'true positive'
ACCURACY_TRUE_NEGATIVE = 'true negative'
ACCURACY_FALSE_POSITIVE = 'false positive'
ACCURACY_FALSE_NEGATIVE = 'false negative'


class DetectionAccuracy(DetectionErrorRate):
    """Detection accuracy

    This metric can be used to evaluate binary classification tasks such as
    speech activity detection, for instance. It can be applied on Timeline
    or Annotation instances. Inputs are expected to only contain segments (or
    tracks) corresponding to the positive class (e.g. speech regions). Gaps in
    the inputs considered as the negative class (e.g. non-speech regions).

    It is computed as (tp + tn) / total, where tp is the duration of true
    positive (e.g. speech classified as speech), tn is the duration of true
    negative (e.g. non-speech classified as non-speech), and total is the total
    duration of the input signal.

    Parameters
    ----------
    collar : float, optional
        Duration (in seconds) of collars removed from evaluation around
        boundaries of reference segments (one half before, one half after).
    """

    @classmethod
    def metric_name(cls):
        return ACCURACY_NAME

    @classmethod
    def metric_components(cls):
        return [ACCURACY_TRUE_POSITIVE, ACCURACY_TRUE_NEGATIVE,
                ACCURACY_FALSE_POSITIVE, ACCURACY_FALSE_NEGATIVE]

    def _get_details(self, reference, hypothesis, uem=None, **kwargs):

        detail = self._init_details()

        confusion = self._confusion(reference, hypothesis, uem=uem)

        try:
            true_positive = confusion.loc['+', '+'].item()
        except KeyError as e:
            true_positive = 0.

        try:
            false_positive = confusion.loc['-', '+'].item()
        except KeyError as e:
            false_positive = 0.

        try:
            true_negative = confusion.loc['-', '-'].item()
        except KeyError as e:
            true_negative = 0.

        try:
            false_negative = confusion.loc['+', '-'].item()
        except KeyError as e:
            false_negative = 0.

        detail[ACCURACY_TRUE_NEGATIVE] = true_negative
        detail[ACCURACY_TRUE_POSITIVE] = true_positive
        detail[ACCURACY_FALSE_NEGATIVE] = false_negative
        detail[ACCURACY_FALSE_POSITIVE] = false_positive

        return detail

    def _get_rate(self, detail):
        numerator = 1. * (detail[ACCURACY_TRUE_NEGATIVE] +
                          detail[ACCURACY_TRUE_POSITIVE])
        denominator = 1. * (detail[ACCURACY_TRUE_NEGATIVE] +
                            detail[ACCURACY_TRUE_POSITIVE] +
                            detail[ACCURACY_FALSE_NEGATIVE] +
                            detail[ACCURACY_FALSE_POSITIVE])

        if denominator == 0.:
            return 1.
        else:
            return numerator / denominator

    def _pretty(self, detail):
        raise NotImplementedError('')

PRECISION_NAME = 'detection precision'
PRECISION_RETRIEVED = 'retrieved'
PRECISION_RELEVANT_RETRIEVED = 'relevant retrieved'


class DetectionPrecision(DetectionErrorRate):
    """Detection precision

    This metric can be used to evaluate binary classification tasks such as
    speech activity detection, for instance. It can be applied on Timeline
    or Annotation instances. Inputs are expected to only contain segments (or
    tracks) corresponding to the positive class (e.g. speech regions). Gaps in
    the inputs considered as the negative class (e.g. non-speech regions).

    It is computed as tp / (tp + fp), where tp is the duration of true positive
    (e.g. speech classified as speech), and fp is the duration of false
    positive (e.g. non-speech classified as speech).

    Parameters
    ----------
    collar : float, optional
        Duration (in seconds) of collars removed from evaluation around
        boundaries of reference segments (one half before, one half after).
    """

    @classmethod
    def metric_name(cls):
        return PRECISION_NAME

    @classmethod
    def metric_components(cls):
        return [PRECISION_RETRIEVED, PRECISION_RELEVANT_RETRIEVED]

    def _get_details(self, reference, hypothesis, uem=None, **kwargs):

        detail = self._init_details()

        confusion = self._confusion(reference, hypothesis, uem=uem)

        try:
            true_positive = confusion.loc['+', '+'].item()
        except KeyError as e:
            true_positive = 0.

        try:
            false_positive = confusion.loc['-', '+'].item()
        except KeyError as e:
            false_positive = 0.

        detail[PRECISION_RETRIEVED] = true_positive + false_positive
        detail[PRECISION_RELEVANT_RETRIEVED] = true_positive

        return detail

    def _get_rate(self, detail):
        relevant_retrieved = 1. * detail[PRECISION_RELEVANT_RETRIEVED]
        retrieved = 1. * detail[PRECISION_RETRIEVED]
        if retrieved == 0.:
            return 1.
        else:
            return relevant_retrieved / retrieved

    def _pretty(self, detail):
        raise NotImplementedError('')

RECALL_NAME = 'detection recall'
RECALL_RELEVANT = 'relevant'
RECALL_RELEVANT_RETRIEVED = 'relevant retrieved'


class DetectionRecall(DetectionErrorRate):
    """Detection recall

    This metric can be used to evaluate binary classification tasks such as
    speech activity detection, for instance. It can be applied on Timeline
    or Annotation instances. Inputs are expected to only contain segments (or
    tracks) corresponding to the positive class (e.g. speech regions). Gaps in
    the inputs considered as the negative class (e.g. non-speech regions).

    It is computed as tp / (tp + fn), where tp is the duration of true positive
    (e.g. speech classified as speech), and fn is the duration of false
    negative (e.g. speech classified as non-speech).

    Parameters
    ----------
    collar : float, optional
        Duration (in seconds) of collars removed from evaluation around
        boundaries of reference segments (one half before, one half after).
    """

    @classmethod
    def metric_name(cls):
        return RECALL_NAME

    @classmethod
    def metric_components(cls):
        return [RECALL_RELEVANT, RECALL_RELEVANT_RETRIEVED]

    def _get_details(self, reference, hypothesis, uem=None, **kwargs):

        detail = self._init_details()
        confusion = self._confusion(reference, hypothesis, uem=uem)

        try:
            true_positive = confusion.loc['+', '+'].item()
        except KeyError as e:
            true_positive = 0.

        try:
            false_negative = confusion.loc['+', '-'].item()
        except KeyError as e:
            false_negative = 0.

        detail[RECALL_RELEVANT] = true_positive + false_negative
        detail[RECALL_RELEVANT_RETRIEVED] = true_positive

        return detail

    def _get_rate(self, detail):
        relevant_retrieved = 1. * detail[RECALL_RELEVANT_RETRIEVED]
        relevant = 1. * detail[RECALL_RELEVANT]
        if relevant == 0.:
            if relevant_retrieved == 0:
                return 1.
            else:
                return 0.
        else:
            return relevant_retrieved / relevant

    def _pretty(self, detail):
        raise NotImplementedError('')


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2014 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""
Evaluation metrics

Usage:
  eval identification [options] <references.mdtm> <hypothesis.mdtm>
  eval -h | --help
  eval --version

Options:
  --uem <eval.uem>         Evaluation map.
  --uris <uris.lst>        List of resources.
  -h --help                Show this screen.
  --version                Show version.
"""

import pyannote.metrics
from pyannote.metrics.identification import IdentificationErrorRate
from pyannote.parser.util import CoParser
from docopt import docopt
import sys


def do_identification(
    references_mdtm, hypothesis_mdtm, uris_lst=None, eval_uem=None
):

    ier = IdentificationErrorRate()

    iter_over = {
        'reference': references_mdtm,
        'hypothesis': hypothesis_mdtm
    }

    if uris_lst:
        iter_over['uris'] = uris_lst
    else:
        iter_over['uris'] = 'reference'

    if eval_uem:
        iter_over['uem'] = eval_uem

    coParser = CoParser(**iter_over)

    for uri, ref, hyp, uem in coParser.iter('uris', 'reference', 'hypothesis',
                                            'uem'):
        rate = ier(ref, hyp, uem=uem)
        sys.stdout.write('{uri:s}: {ier:3.2f}%\n'.format(uri=uri, ier=100 * rate))
        sys.stdout.flush()

    sys.stdout.write('Total: {ier:3.2f}%'.format(ier=100 * abs(ier)))

if __name__ == '__main__':

    version = 'PyAnnoteEval v{version:s}'.format(
        version=pyannote.metrics.__version__)
    arguments = docopt(__doc__, version=version)

    if arguments['identification']:
        references_mdtm = arguments['<references.mdtm>']
        hypothesis_mdtm = arguments['<hypothesis.mdtm>']
        uris_lst = arguments['--uris']
        eval_uem = arguments['--uem']

        do_identification(references_mdtm, hypothesis_mdtm,
                          uris_lst=uris_lst, eval_uem=eval_uem)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2012-2016 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from __future__ import unicode_literals

import scipy.stats
import numpy as np
from pyannote.algorithms.tagging import DirectTagger


class BaseMetric(object):
    """
    :class:`BaseMetric` is the base class for most PyAnnote evaluation metrics.

    Parameters
    ----------
    name : str
        Human-readable name of the metric (eg. 'diarization error rate')
    components : list, set or tuple
        Human-readable names of the components of the metric
        (eg. ['correct', 'false alarm', 'miss', 'confusion'])

    """

    @classmethod
    def metric_name(cls):
        raise NotImplementedError("Missing class method 'metric_name'.")

    @classmethod
    def metric_components(cls):
        raise NotImplementedError("Missing class method 'metric_components'")

    def __init__(self, **kwargs):
        super(BaseMetric, self).__init__()
        self.__name = self.__class__.metric_name()
        self.__values = set(self.__class__.metric_components())
        self.reset()
        self._tagger = DirectTagger()

    def __get_name(self):
        return self.__class__.metric_name()
    name = property(fget=__get_name, doc="Metric name.")

    def __accumulate(self, components):
        """Accumulate metric components

        Parameters
        ----------
        components : dict
            Dictionary where keys are component names and values are component
            values

        """
        for name in self.__values:
            self.__details[name] += components[name]

    def __compute(self, components, accumulate=True, detailed=False):
        """Compute metric value from computed `components`

        Parameters
        ----------
        components : dict
            Dictionary where keys are components names and values are component
            values
        accumulate : bool, optional
            If True, components are accumulated. Defaults to True.
        detailed : bool, optional
            By default (False), return metric value only.
            Set `detailed` to True to return `components` updated with metric
            value (for key `self.name`).

        Returns
        -------
        value (if `detailed` is False) : float
            Metric value
        components (if `detailed` is True) : dict
            `components` updated with metric value
        """
        if accumulate:
            self.__accumulate(components)
        rate = self._get_rate(components)
        if detailed:
            components = dict(components)
            components.update({self.__name: rate})
            return components
        else:
            return rate

    def __call__(self, reference, hypothesis, detailed=False, **kwargs):
        """Compute metric value and accumulate components

        Parameters
        ----------
        reference : type depends on the metric
            Manual `reference`
        hypothesis : same as `reference`
            Evaluated `hypothesis`
        detailed : bool, optional
            By default (False), return metric value only.

            Set `detailed` to True to return dictionary where keys are
            components names and values are component values


        Returns
        -------
        value : float (if `detailed` is False)
            Metric value
        components : dict (if `detailed` is True)
            `components` updated with metric value

        """
        detail = self._get_details(reference, hypothesis, **kwargs)
        self.__rates.append((reference.uri, self._get_rate(detail)))
        return self.__compute(detail, accumulate=True, detailed=detailed)

    def __str__(self):
        detail = self.__compute(self.__details,
                                accumulate=False,
                                detailed=True)
        return self._pretty(detail)

    def __abs__(self):
        """Compute metric value from accumulated components"""
        return self._get_rate(self.__details)

    def __getitem__(self, component):
        """Get value of accumulated `component`.

        Parameters
        ----------
        component : str
            Name of `component`

        Returns
        -------
        value : type depends on the metric
            Value of accumulated `component`

        """
        if component == slice(None, None, None):
            return dict(self.__details)
        else:
            return self.__details[component]

    def __iter__(self):
        """Iterator over the accumulated (uri, value)"""
        for v, r in self.__rates:
            yield v, r

    def _get_details(self, reference, hypothesis, **kwargs):
        """Compute metric components

        Parameters
        ----------
        reference : type depends on the metric
            Manual `reference`
        hypothesis : same as `reference`
            Evaluated `hypothesis`

        Returns
        -------
        components : dict
            Dictionary where keys are component names and values are component
            values

        """
        raise NotImplementedError("Missing method '_get_details'.")

    def _get_rate(self, components):
        """Compute metric value from computed `components`

        Parameters
        ----------
        components : dict
            Dictionary where keys are components names and values are component
            values

        Returns
        -------
        value : type depends on the metric
            Metric value
        """
        raise NotImplementedError("Missing method '_get_rate'.")

    def _pretty(self, components):
        string = '%s: %g\n' % (self.name, components[self.name])
        for name, value in components.items():
            if name == self.name:
                continue
            string += '  - %s: %g\n' % (name, value)
        return string

    def _init_details(self):
        return {value: 0. for value in self.__values}

    def reset(self):
        """Reset accumulated components and metric values"""
        self.__details = self._init_details()
        self.__rates = []

    def confidence_interval(self, alpha=0.9):
        """Compute confidence interval on accumulated metric values

        Parameters
        ----------
        alpha : float, optional
            Probability that the returned confidence interval contains
            the true metric value.

        Returns
        -------
        (center, (lower, upper))
            with center the mean of the conditional pdf of the metric value
            and (lower, upper) is a confidence interval centered on the median,
            containing the estimate to a probability alpha.

        See Also:
        ---------
        scipy.stats.bayes_mvs

        """
        m, _, _ = scipy.stats.bayes_mvs([r for _, r in self.__rates], alpha=alpha)
        return m


PRECISION_NAME = 'precision'
PRECISION_RETRIEVED = '# retrieved'
PRECISION_RELEVANT_RETRIEVED = '# relevant retrieved'


class Precision(BaseMetric):
    """
    :class:`Precision` is a base class for precision-like evaluation metrics.

    It defines two components '# retrieved' and '# relevant retrieved' and the
    _get_rate() method to compute the actual precision:

        Precision = # retrieved / # relevant retrieved

    Inheriting classes must implement _get_details().
    """

    @classmethod
    def metric_name(cls):
        return PRECISION_NAME

    @classmethod
    def metric_components(cls):
        return [PRECISION_RETRIEVED, PRECISION_RELEVANT_RETRIEVED]

    def _get_rate(self, components):
        """Compute precision from `components`"""
        numerator = components[PRECISION_RELEVANT_RETRIEVED]
        denominator = components[PRECISION_RETRIEVED]
        if denominator == 0.:
            if numerator == 0:
                return 1.
            else:
                raise ValueError('')
        else:
            return numerator/denominator

RECALL_NAME = 'recall'
RECALL_RELEVANT = '# relevant'
RECALL_RELEVANT_RETRIEVED = '# relevant retrieved'


class Recall(BaseMetric):
    """
    :class:`Recall` is a base class for recall-like evaluation metrics.

    It defines two components '# relevant' and '# relevant retrieved' and the
    _get_rate() method to compute the actual recall:

        Recall = # relevant retrieved / # relevant

    Inheriting classes must implement _get_details().
    """

    @classmethod
    def metric_name(cls):
        return RECALL_NAME

    @classmethod
    def metric_components(cls):
        return [RECALL_RELEVANT, RECALL_RELEVANT_RETRIEVED]

    def _get_rate(self, components):
        """Compute recall from `components`"""
        numerator = components[RECALL_RELEVANT_RETRIEVED]
        denominator = components[RECALL_RELEVANT]
        if denominator == 0.:
            if numerator == 0:
                return 1.
            else:
                raise ValueError('')
        else:
            return numerator/denominator


def f_measure(precision, recall, beta=1.):
    """Compute f-measure

    f-measure is defined as follows:
        F(P, R, b) = (1+b²).P.R / (b².P + R)

    where P is `precision`, R is `recall` and b is `beta`
    """
    return (1+beta*beta)*precision*recall / (beta*beta*precision+recall)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2012-2016 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from __future__ import unicode_literals

"""Metrics for diarization"""

import numpy as np
from xarray import DataArray

from pyannote.algorithms.tagging.mapping import HungarianMapper
from pyannote.algorithms.tagging.mapping import GreedyMapper

from .base import BaseMetric
from .identification import IdentificationErrorRate

DER_NAME = 'diarization error rate'


class DiarizationErrorRate(IdentificationErrorRate):
    """Diarization error rate

    First, the optimal mapping between reference and hypothesis labels
    is obtained using the Hungarian algorithm. Then, the actual diarization
    error rate is computed as the identification error rate with each hypothesis
    label translated into the corresponding reference label.

    * Diarization error rate between `reference` and `hypothesis` annotations

        >>> metric = DiarizationErrorRate()
        >>> reference = Annotation(...)           # doctest: +SKIP
        >>> hypothesis = Annotation(...)          # doctest: +SKIP
        >>> value = metric(reference, hypothesis) # doctest: +SKIP

    * Compute global diarization error rate and confidence interval
      over multiple documents

        >>> for reference, hypothesis in ...      # doctest: +SKIP
        ...    metric(reference, hypothesis)      # doctest: +SKIP
        >>> global_value = abs(metric)            # doctest: +SKIP
        >>> mean, (lower, upper) = metric.confidence_interval() # doctest: +SKIP

    * Get diarization error rate detailed components

        >>> components = metric(reference, hypothesis, detailed=True) #doctest +SKIP

    * Get accumulated components

        >>> components = metric[:]                # doctest: +SKIP
        >>> metric['confusion']                   # doctest: +SKIP

    See Also
    --------
    :class:`pyannote.metric.base.BaseMetric`: details on accumulation
    :class:`pyannote.metric.identification.IdentificationErrorRate`: identification error rate

    """

    @classmethod
    def metric_name(cls):
        return DER_NAME

    def __init__(self, **kwargs):
        super(DiarizationErrorRate, self).__init__()
        self._mapper = HungarianMapper()

    def optimal_mapping(self, reference, hypothesis):
        """Optimal label mapping"""
        return self._mapper(hypothesis, reference)

    def _get_details(self, reference, hypothesis, **kwargs):
        reference = reference.anonymize_labels(generator='string')
        hypothesis = hypothesis.anonymize_labels(generator='int')
        mapping = self.optimal_mapping(reference, hypothesis)
        return super(DiarizationErrorRate, self)\
            ._get_details(reference, hypothesis.translate(mapping))


class GreedyDiarizationErrorRate(IdentificationErrorRate):
    """Greedy diarization error rate

    First, the greedy mapping between reference and hypothesis labels is
    obtained. Then, the actual diarization error rate is computed as the
    identification error rate with each hypothesis label translated into the
    corresponding reference label.

    * Greedy diarization error rate between `reference` and `hypothesis` annotations

        >>> metric = GreedyDiarizationErrorRate()
        >>> reference = Annotation(...)           # doctest: +SKIP
        >>> hypothesis = Annotation(...)          # doctest: +SKIP
        >>> value = metric(reference, hypothesis) # doctest: +SKIP

    * Compute global greedy diarization error rate and confidence interval
      over multiple documents

        >>> for reference, hypothesis in ...      # doctest: +SKIP
        ...    metric(reference, hypothesis)      # doctest: +SKIP
        >>> global_value = abs(metric)            # doctest: +SKIP
        >>> mean, (lower, upper) = metric.confidence_interval() # doctest: +SKIP

    * Get greedy diarization error rate detailed components

        >>> components = metric(reference, hypothesis, detailed=True) #doctest +SKIP

    * Get accumulated components

        >>> components = metric[:]                # doctest: +SKIP
        >>> metric['confusion']                   # doctest: +SKIP

    See Also
    --------
    :class:`pyannote.metric.base.BaseMetric`: details on accumulation

    """

    @classmethod
    def metric_name(cls):
        return DER_NAME

    def __init__(self, **kwargs):
        super(GreedyDiarizationErrorRate, self).__init__()
        self._mapper = GreedyMapper()

    def greedy_mapping(self, reference, hypothesis):
        """Greedy label mapping"""
        return self._mapper(hypothesis, reference)

    def _get_details(self, reference, hypothesis, **kwargs):
        reference = reference.anonymize_labels(generator='string')
        hypothesis = hypothesis.anonymize_labels(generator='int')
        mapping = self.greedy_mapping(reference, hypothesis)
        return super(GreedyDiarizationErrorRate, self)\
            ._get_details(reference, hypothesis.translate(mapping))


PURITY_NAME = 'purity'
PURITY_TOTAL = 'total'
PURITY_CORRECT = 'correct'


class DiarizationPurity(BaseMetric):
    """Purity

    Compute purity of hypothesis clusters with respect to reference classes.

    Parameters
    ----------
    per_cluster : bool, optional
        By default (per_cluster = False), clusters are duration-weighted.
        When per_cluster = True, each cluster is given the same weight.

    """

    @classmethod
    def metric_name(cls):
        return PURITY_NAME

    @classmethod
    def metric_components(cls):
        return [PURITY_TOTAL, PURITY_CORRECT]

    def __init__(self, detection_error=False, per_cluster=False, **kwargs):
        super(DiarizationPurity, self).__init__()
        self.per_cluster = per_cluster

    def _get_details(self, reference, hypothesis, **kwargs):
        detail = self._init_details()

        matrix = reference.smooth() * hypothesis.smooth()

        biggest = matrix.max(dim='i')
        duration = DataArray(
            [hypothesis.label_duration(j.item()) for j in biggest.coords['j']],
            coords=biggest.coords)

        if self.per_cluster:
            # biggest class in each cluster
            detail[PURITY_CORRECT] = (biggest / duration).sum().item()
            # number of clusters (as float)
            detail[PURITY_TOTAL] = len(biggest)

        else:
            if np.prod(matrix.shape):
                detail[PURITY_CORRECT] = biggest.sum().item()
            else:
                detail[PURITY_CORRECT] = 0.
            # total duration of clusters (with overlap)
            detail[PURITY_TOTAL] = duration.sum().item()

        return detail

    def _get_rate(self, detail):
        if detail[PURITY_TOTAL] > 0.:
            return detail[PURITY_CORRECT] / detail[PURITY_TOTAL]
        else:
            return 1.

    def _pretty(self, detail):
        string = ""
        if self.per_cluster:
            string += "  - clusters: %d\n" % (detail[PURITY_TOTAL])
            string += "  - correct: %.2f\n" % (detail[PURITY_CORRECT])
        else:
            string += "  - duration: %.2f seconds\n" % (detail[PURITY_TOTAL])
            string += "  - correct: %.2f seconds\n" % (detail[PURITY_CORRECT])
        string += "  - %s: %.2f %%\n" % (self.name, 100*detail[self.name])
        return string

COVERAGE_NAME = 'coverage'


class DiarizationCoverage(DiarizationPurity):
    """Coverage

    Compute coverage of hypothesis clusters with respect to reference classes
    (i.e. purity of reference classes with respect to hypothesis clusters)

    Parameters
    ----------
    per_cluster : bool, optional
        By default (per_cluster = False), classes are duration-weighted.
        When per_cluster = True, each class is given the same weight.
    """

    @classmethod
    def metric_name(cls):
        return COVERAGE_NAME

    def __init__(self, per_cluster=False, **kwargs):
        super(DiarizationCoverage, self).__init__(per_cluster=per_cluster)

    def _get_details(self, reference, hypothesis, **kwargs):
        return super(DiarizationCoverage, self)\
            ._get_details(hypothesis, reference)

    def _pretty(self, detail):
        string = ""
        if self.per_cluster:
            string += "  - classes: %d\n" % (detail[PURITY_TOTAL])
            string += "  - correct: %.2f\n" % (detail[PURITY_CORRECT])
        else:
            string += "  - duration: %.2f seconds\n" % (detail[PURITY_TOTAL])
            string += "  - correct: %.2f seconds\n" % (detail[PURITY_CORRECT])
        string += "  - %s: %.2f %%\n" % (self.name, 100*detail[self.name])
        return string


HOMOGENEITY_NAME = 'homogeneity'
HOMOGENEITY_ENTROPY = 'entropy'
HOMOGENEITY_CROSS_ENTROPY = 'cross-entropy'


class DiarizationHomogeneity(BaseMetric):
    """Homogeneity"""

    @classmethod
    def metric_name(cls):
        return HOMOGENEITY_NAME

    @classmethod
    def metric_components(cls):
        return [HOMOGENEITY_ENTROPY, HOMOGENEITY_CROSS_ENTROPY]

    def _get_details(self, reference, hypothesis, **kwargs):
        detail = self._init_details()

        matrix = reference.smooth() * hypothesis.smooth()

        duration = matrix.sum()
        rduration = matrix.sum(dim='j')
        hduration = matrix.sum(dim='i')

        # Reference entropy and reference/hypothesis cross-entropy
        ratio = rduration / duration
        entropy = -(ratio * np.log(ratio)).sum()
        detail[HOMOGENEITY_ENTROPY] = entropy.item()

        cross_entropy = -((matrix / duration) * np.log(matrix / hduration)).sum()
        detail[HOMOGENEITY_CROSS_ENTROPY] = cross_entropy.item()

        return detail

    def _get_rate(self, detail):
        numerator = 1. * detail[HOMOGENEITY_CROSS_ENTROPY]
        denominator = 1. * detail[HOMOGENEITY_ENTROPY]
        if denominator == 0.:
            if numerator == 0:
                return 1.
            else:
                return 0.
        else:
            return 1. - numerator/denominator

    def _pretty(self, detail):
        string = ""
        string += "  - %s: %.2f\n" % \
                  (HOMOGENEITY_ENTROPY, detail[HOMOGENEITY_ENTROPY])
        string += "  - %s: %.2f\n" % \
                  (HOMOGENEITY_CROSS_ENTROPY, detail[HOMOGENEITY_CROSS_ENTROPY])
        string += "  - %s: %.2f %%\n" % (self.name, 100*detail[self.name])
        return string

COMPLETENESS_NAME = 'completeness'


class DiarizationCompleteness(DiarizationHomogeneity):
    """Completeness"""

    @classmethod
    def metric_name(cls):
        return COMPLETENESS_NAME

    def _get_details(self, reference, hypothesis, **kwargs):
        return super(DiarizationCompleteness, self)\
            ._get_details(hypothesis, reference)


if __name__ == "__main__":
    import doctest
    doctest.testmod()
